@@ -3,8 +3,8 @@
 /*
     This file is part of Corrade.
 
-    Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016
-              Vladimír Vondruš <mosra@centrum.cz>
+    Copyright © 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2016,
+                2017, 2018 Vladimír Vondruš <mosra@centrum.cz>
 
     Permission is hereby granted, free of charge, to any person obtaining a
     copy of this software and associated documentation files (the "Software"),
@@ -43,6 +43,25 @@
 namespace Corrade { namespace Containers {
 
 namespace Implementation {
+    template<class T> struct DefaultDeleter {
+        T operator()() const { return T{}; }
+    };
+    template<class T> struct DefaultDeleter<void(*)(T*, std::size_t)> {
+        void(*operator()() const)(T*, std::size_t) { return nullptr; }
+    };
+
+    template<class T, class D> struct CallDeleter {
+        void operator()(D deleter, T* data, std::size_t size) const {
+            deleter(data, size);
+        }
+    };
+    template<class T> struct CallDeleter<T, void(*)(T*, std::size_t)> {
+        void operator()(void(*deleter)(T*, std::size_t), T* data, std::size_t size) const {
+            if(deleter) deleter(data, size);
+            else delete[] data;
+        }
+    };
+
     template<class T> void noInitDeleter(T* data, std::size_t size) {
         if(data) for(T *it = data, *end = data + size; it != end; ++it)
             it->~T();
@@ -50,34 +69,23 @@ namespace Implementation {
     }
 }
 
-/**
-@brief Default array deleter
-@param data     Array pointer
-@param size     Array element count
-
-Equivalent to calling `delete[]` on passed pointer, @p size is ignored.
-*/
-template<class T> void defaultDeleter(T* data, std::size_t size) {
-    static_cast<void>(size);
-    delete[] data;
-}
-
 /**
 @brief Array wrapper with size information
 @tparam T   Element type
-@tparam D   Deleter type, defaults to pointer to function of the same signature
-    as @ref defaultDeleter()
+@tparam D   Deleter type. Defaults to pointer to @cpp void(T*, std::size_t) @ce
+    function, where first is array pointer and second array size
 
 Provides movable RAII wrapper around plain C array. Main use case is storing
 binary data of unspecified type, where addition/removal of elements is not
 needed or harmful.
 
-However, the class is usable also as lighter non-copyable alternative to
-`std::vector`, in STL algorithms in the same way as plain C array and
-additionally also in range-based for cycle.
+However, the class is usable also as a lighter non-copyable alternative to
+@ref std::vector; usable in STL algorithms in the same way as plain C array and
+additionally also in range-for cycle.
 
 Usage example:
-@code
+
+@code{.cpp}
 // Create default-initialized array with 5 integers and set them to some value
 Containers::Array<int> a{5};
 int b = 0;
@@ -88,7 +96,7 @@ auto b = Containers::Array<int>::from(3, 18, -157, 0);
 b[3] = 25; // b = {3, 18, -157, 25}
 @endcode
 
-## Array initialization
+@section Containers-Array-initialization Array initialization
 
 The array is by default *default-initialized*, which means that trivial types
 are not initialized at all and default constructor is called on other types. It
@@ -98,14 +106,15 @@ is possible to initialize the array in a different way using so-called *tags*:
     (useful when you want to make the choice appear explicit).
 -   @ref Array(ValueInitT, std::size_t) zero-initializes trivial types and
     calls default constructor elsewhere.
--   @ref Array(DirectInitT, std::size_t, Args...) constructs all elements of
-    the array using provided arguments.
+-   @ref Array(DirectInitT, std::size_t, Args&&... args) constructs all
+    elements of the array using provided arguments.
 -   @ref Array(NoInitT, std::size_t) does not initialize anything and you need
     to call the constructor on all elements manually using placement new,
-    `std::uninitialized_copy` or similar. This is the dangerous option.
+    @ref std::uninitialized_copy() or similar. This is the dangerous option.
 
 Example:
-@code
+
+@code{.cpp}
 // These are equivalent
 Containers::Array<int> a1{5};
 Containers::Array<int> a2{Containers::DefaultInit, 5};
@@ -128,19 +137,20 @@ int index = 0;
 for(Foo& f: d) new(&f) Foo(index++);
 @endcode
 
-## Wrapping externally allocated arrays
+@section Containers-Array-wrapping Wrapping externally allocated arrays
 
-By default the class makes all allocations using `operator new[]` and
-deallocates using `operator delete[]` for given @p T, with some additional
-trickery done internally to make the @ref Array(NoInitT, std::size_t) and
-@ref Array(DirectInitT, std::size_t, ...) constructors work. When wrapping an
-externally allocated array using @ref Array(T*, std::size_t, D), it is possible
-to specify which function to use for deallocation. By default the
-@ref defaultDeleter() function is used, which is equivalent to the
-`operator delete[]`.
+By default the class makes all allocations using @cpp operator new[] @ce and
+deallocates using @cpp operator delete[] @ce for given @p T, with some
+additional trickery done internally to make the @ref Array(NoInitT, std::size_t)
+and @ref Array(DirectInitT, std::size_t, Args&&... args) constructors work.
+When wrapping an externally allocated array using @ref Array(T*, std::size_t, D),
+it is possible to specify which function to use for deallocation. By default
+the deleter is set to @cpp nullptr @ce, which is equivalent to deleting the
+contents using @cpp operator delete[] @ce.
 
-For example, properly deallocating array allocated using `std::malloc()`:
-@code
+For example, properly deallocating array allocated using @ref std::malloc():
+
+@code{.cpp}
 const int* data = reinterpret_cast<int*>(std::malloc(25*sizeof(int)));
 
 // Will call std::free() on destruction
@@ -150,7 +160,8 @@ Containers::Array<int> array{data, 25, [](int* data, std::size_t) { std::free(da
 By default, plain function pointers are used to avoid having the type affected
 by the deleter function. If the deleter needs to manage some state, a custom
 deleter type can be used:
-@code
+
+@code{.cpp}
 struct UnmapBuffer {
     UnmapBuffer(GLuint id): _id{id} {}
     void operator()(T*, std::size_t) { glUnmapNamedBuffer(_id); }
@@ -166,6 +177,8 @@ char* data = reinterpret_cast<char*>(glMapNamedBuffer(buffer, GL_READ_WRITE));
 Containers::Array<char, UnmapBuffer> array{data, bufferSize, UnmapBuffer{buffer}};
 @endcode
 
+@see @ref arrayCast(Array<T, D>&)
+
 @todo Something like ArrayTuple to create more than one array with single
     allocation and proper alignment for each type? How would non-POD types be
     constructed in that? Will that be useful in more than one place?
@@ -180,18 +193,17 @@ class Array {
         typedef T Type;     /**< @brief Element type */
         typedef D Deleter;  /**< @brief Deleter type */
 
-        /**
-         * @brief Create array from given values
-         *
-         * Zero argument count creates `nullptr` array.
-         */
-        template<class ...U> static Array<T, D> from(U&&... values) {
-            return fromInternal(std::forward<U>(values)...);
-        }
-
         #ifdef CORRADE_BUILD_DEPRECATED
         /**
-         * @copybrief Array(ValueInitT, std::size_t)
+         * @brief @copybrief Array(InPlaceInitT, std::initializer_list<T>)
+         * @deprecated Use @ref Array(InPlaceInitT, std::initializer_list<T>) instead.
+         */
+        template<class ...U> CORRADE_DEPRECATED("use Array(InPlaceInitT, std::initializer_list<T>) instead") static Array<T, D> from(U&&... values) {
+            return Array<T, D>{InPlaceInit, {T(std::forward<U>(values))...}};
+        }
+
+        /**
+         * @brief @copybrief Array(ValueInitT, std::size_t)
          * @deprecated Use @ref Array(ValueInitT, std::size_t) instead.
          */
         CORRADE_DEPRECATED("use Array(ValueInitT, std::size_t) instead") static Array<T, D> zeroInitialized(std::size_t size) {
@@ -205,7 +217,8 @@ class Array {
         #else
         template<class U, class V = typename std::enable_if<std::is_same<std::nullptr_t, U>::value>::type> /*implicit*/ Array(U) noexcept:
         #endif
-            _data{nullptr}, _size{0}, _deleter{defaultDeleter} {}
+            /* GCC <=4.8 breaks on _deleter{} */
+            _data{nullptr}, _size{0}, _deleter(Implementation::DefaultDeleter<D>{}()) {}
 
         /**
          * @brief Default constructor
@@ -213,7 +226,8 @@ class Array {
          * Creates zero-sized array. Move array with nonzero size onto the
          * instance to make it useful.
          */
-        /*implicit*/ Array() noexcept: _data(nullptr), _size(0), _deleter{defaultDeleter} {}
+        /* GCC <=4.8 breaks on _deleter{} */
+        /*implicit*/ Array() noexcept: _data(nullptr), _size(0), _deleter(Implementation::DefaultDeleter<D>{}()) {}
 
         /**
          * @brief Construct default-initialized array
@@ -223,7 +237,7 @@ class Array {
          * allocation is done.
          * @see @ref DefaultInit, @ref Array(ValueInitT, std::size_t)
          */
-        explicit Array(DefaultInitT, std::size_t size): _data{size ? new T[size] : nullptr}, _size{size}, _deleter{defaultDeleter} {}
+        explicit Array(DefaultInitT, std::size_t size): _data{size ? new T[size] : nullptr}, _size{size}, _deleter{nullptr} {}
 
         /**
          * @brief Construct value-initialized array
@@ -237,7 +251,7 @@ class Array {
          * them to zero.
          * @see @ref ValueInit, @ref Array(DefaultInitT, std::size_t)
          */
-        explicit Array(ValueInitT, std::size_t size): _data{size ? new T[size]() : nullptr}, _size{size}, _deleter{defaultDeleter} {}
+        explicit Array(ValueInitT, std::size_t size): _data{size ? new T[size]() : nullptr}, _size{size}, _deleter{nullptr} {}
 
         /**
          * @brief Construct the array without initializing its contents
@@ -247,12 +261,12 @@ class Array {
          * placement new.
          *
          * Useful if you will be overwriting all elements later anyway.
-         * @attention Internally the data are allocated as `char` array and
-         *      destruction is done using custom deleter that explicitly calls
-         *      destructor on *all elements* regardless of whether they were
-         *      properly constructed or not and then deallocates the data as
-         *      `char` array.
-         * @see @ref NoInit, @ref Array(DirectInitT, std::size_t, Args...),
+         * @attention Internally the data are allocated as @cpp char @ce array
+         *      and destruction is done using custom deleter that explicitly
+         *      calls destructor on *all elements* regardless of whether they
+         *      were properly constructed or not and then deallocates the data
+         *      as @cpp char @ce array.
+         * @see @ref NoInit, @ref Array(DirectInitT, std::size_t, Args&&... args),
          *      @ref deleter()
          */
         explicit Array(NoInitT, std::size_t size): _data{size ? reinterpret_cast<T*>(new char[size*sizeof(T)]) : nullptr}, _size{size}, _deleter{Implementation::noInitDeleter} {}
@@ -262,10 +276,22 @@ class Array {
          *
          * Allocates the array using the @ref Array(NoInitT, std::size_t)
          * constructor and then initializes each element with placement new
-         * using forwarded @p arguments.
+         * using forwarded @p args.
          */
         template<class... Args> explicit Array(DirectInitT, std::size_t size, Args&&... args);
 
+        /**
+         * @brief Construct list-initialized array
+         *
+         * Allocates the array using the @ref Array(NoInitT, std::size_t)
+         * constructor and then copy-initializes each element with placement
+         * new using values from @p list.
+         */
+        /* There is no initializer-list constructor because it would make
+           Containers::Array<std::size_t>{5} behave differently (and that would
+           break things *badly* as this is a *very* common use-case) */
+        explicit Array(InPlaceInitT, std::initializer_list<T> list);
+
         /**
          * @brief Construct default-initialized array
          *
@@ -281,9 +307,10 @@ class Array {
          * @p deleter. See class documentation for more information about
          * custom deleters and @ref ArrayView for non-owning array wrapper.
          */
-        explicit Array(T* data, std::size_t size, D deleter = defaultDeleter): _data{data}, _size{size}, _deleter{deleter} {}
+        /* GCC <=4.8 breaks on _deleter{} */
+        explicit Array(T* data, std::size_t size, D deleter = Implementation::DefaultDeleter<D>{}()): _data{data}, _size{size}, _deleter(deleter) {}
 
-        ~Array() { _deleter(_data, _size); }
+        ~Array() { Implementation::CallDeleter<T, D>{}(_deleter, _data, _size); }
 
         /** @brief Copying is not allowed */
         Array(const Array<T, D>&) = delete;
@@ -297,9 +324,9 @@ class Array {
         /** @brief Move assignment */
         Array<T, D>& operator=(Array<T, D>&&) noexcept;
 
-        #ifndef CORRADE_MSVC2015_COMPATIBILITY
+        #ifndef CORRADE_MSVC2017_COMPATIBILITY
         /** @brief Whether the array is non-empty */
-        /* Disabled on MSVC <= 2015 to avoid ambiguous operator+() when doing
+        /* Disabled on MSVC <= 2017 to avoid ambiguous operator+() when doing
            pointer arithmetic. */
         explicit operator bool() const { return _data; }
         #endif
@@ -307,30 +334,43 @@ class Array {
         /**
          * @brief Convert to @ref ArrayView
          *
-         * Enabled only if `T*` is implicitly convertible to `U*`. Note
-         * that, similarly as with raw pointers, you need to ensure that both
-         * types have the same size.
+         * Enabled only if @cpp T* @ce is implicitly convertible to @cpp U* @ce.
+         * Expects that both types have the same size.
+         * @see @ref arrayView(Array<T, D>&)
          */
         #ifdef DOXYGEN_GENERATING_OUTPUT
         template<class U>
         #else
-        template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value>::type>
+        template<class U, class = typename std::enable_if<!std::is_void<U>::value && std::is_convertible<T*, U*>::value>::type>
         #endif
-        /*implicit*/ operator ArrayView<U>() noexcept { return {_data, _size}; }
+        /*implicit*/ operator ArrayView<U>() noexcept {
+            static_assert(sizeof(T) == sizeof(U), "type sizes are not compatible");
+            return {_data, _size};
+        }
 
         /**
          * @brief Convert to const @ref ArrayView
          *
-         * Enabled only if `const T*` is implicitly convertible to `U*`. Note
-         * that, similarly as with raw pointers, you need to ensure that both
-         * types have the same size.
+         * Enabled only if @cpp T* @ce or @cpp const T* @ce is implicitly
+         * convertible to @cpp U* @ce. Expects that both types have the same
+         * size.
+         * @see @ref arrayView(const Array<T, D>&)
          */
         #ifdef DOXYGEN_GENERATING_OUTPUT
         template<class U>
         #else
-        template<class U, class V = typename std::enable_if<std::is_convertible<T*, U*>::value>::type>
+        template<class U, class = typename std::enable_if<std::is_convertible<T*, U*>::value || std::is_convertible<T*, const U*>::value>::type>
         #endif
-        /*implicit*/ operator ArrayView<const U>() const noexcept { return {_data, _size}; }
+        /*implicit*/ operator ArrayView<const U>() const noexcept {
+            static_assert(sizeof(T) == sizeof(U), "type sizes are not compatible");
+            return {_data, _size};
+        }
+
+        /** @overload */
+        /*implicit*/ operator ArrayView<const void>() const noexcept {
+            /* Yes, the size is properly multiplied by sizeof(T) by the constructor */
+            return {_data, _size};
+        }
 
         /* `char* a = Containers::Array<char>(5); a[3] = 5;` would result in
            instant segfault, disallowing it in the following conversion
@@ -357,7 +397,9 @@ class Array {
         /**
          * @brief Array deleter
          *
-         * @see @ref Array(T*, std::size_t, D), @ref defaultDeleter()
+         * If set to @cpp nullptr @ce, the contents are deleted using standard
+         * @cpp operator delete[] @ce.
+         * @see @ref Array(T*, std::size_t, D)
          */
         D deleter() const { return _deleter; }
 
@@ -401,7 +443,8 @@ class Array {
         /**
          * @brief Fixed-size array slice
          *
-         * Both @p begin and `begin + size` are expected to be in range.
+         * Both @cpp begin @ce and @cpp begin + size @ce are expected to be in
+         * range.
          */
         template<std::size_t size> StaticArrayView<size, T> slice(T* begin) {
             return ArrayView<T>(*this).template slice<size>(begin);
@@ -458,22 +501,70 @@ class Array {
         T* release();
 
     private:
-        template<class ...U> static Array<T, D> fromInternal(U&&... values) {
-            Array<T, D> array;
-            array._size = sizeof...(values);
-            array._data = new T[sizeof...(values)] { T(std::forward<U>(values))... };
-            return array;
-        }
-        /* Specialization for zero argument count */
-        static Array<T, D> fromInternal() { return nullptr; }
-
         T* _data;
         std::size_t _size;
         D _deleter;
 };
 
+/** @relatesalso ArrayView
+@brief Make view on @ref Array
+
+Convenience alternative to calling @ref Array::operator ArrayView<U>()
+explicitly. The following two lines are equivalent:
+
+@code{.cpp}
+Containers::Array<std::uint32_t> data;
+
+Containers::ArrayView<std::uint32_t> a{data};
+auto b = Containers::arrayView(data);
+@endcode
+*/
+template<class T, class D> inline ArrayView<T> arrayView(Array<T, D>& array) {
+    return ArrayView<T>{array};
+}
+
+/** @relatesalso ArrayView
+@brief Make view on const @ref Array
+
+Convenience alternative to calling @ref Array::operator ArrayView<U>()
+explicitly. The following two lines are equivalent:
+
+@code{.cpp}
+const Containers::Array<std::uint32_t> data;
+
+Containers::ArrayView<const std::uint32_t> a{data};
+auto b = Containers::arrayView(data);
+@endcode
+*/
+template<class T, class D> inline ArrayView<const T> arrayView(const Array<T, D>& array) {
+    return ArrayView<const T>{array};
+}
+
+/** @relatesalso Array
+@brief Reinterpret-cast an array
+
+See @ref arrayCast(ArrayView<T>) for more information.
+*/
+template<class U, class T, class D> inline ArrayView<U> arrayCast(Array<T, D>& array) {
+    return arrayCast<U>(arrayView(array));
+}
+
+/** @overload */
+template<class U, class T, class D> inline ArrayView<const U> arrayCast(const Array<T, D>& array) {
+    return arrayCast<const U>(arrayView(array));
+}
+
+/** @relatesalso Array
+@brief Array size
+
+See @ref arraySize(ArrayView<T>) for more information.
+*/
+template<class T> std::size_t arraySize(const Array<T>& view) {
+    return view.size();
+}
+
 #ifdef CORRADE_BUILD_DEPRECATED
-/** @copybrief ArrayView
+/** @brief @copybrief ArrayView
  * @deprecated Use @ref ArrayView.h and @ref ArrayView instead.
  */
 #ifndef CORRADE_MSVC2015_COMPATIBILITY /* Multiple definitions still broken */
@@ -491,6 +582,11 @@ template<class T, class D> template<class ...Args> Array<T, D>::Array(DirectInit
         new(_data + i) T{std::forward<Args>(args)...};
 }
 
+template<class T, class D> Array<T, D>::Array(InPlaceInitT, std::initializer_list<T> list): Array{NoInit, list.size()} {
+    std::size_t i = 0;
+    for(const T& item: list) new(_data + i++) T{item};
+}
+
 template<class T, class D> inline Array<T, D>& Array<T, D>::operator=(Array<T, D>&& other) noexcept {
     using std::swap;
     swap(_data, other._data);
